import json
from collections import defaultdict
from typing import Any, List, Iterable, Union, Optional, Callable


def sort_with_custom_orders(values: List[Any],
                            prefix_orders: Optional[List[Any]] = None,
                            suffix_orders: Optional[List[Any]] = None,
                            key=None, hash_fun=None, reverse=False) -> List[Any]:
    class Empty:
        pass

    UNSIGNED = Empty()

    if prefix_orders is None:
        prefix_orders = []

    if suffix_orders is None:
        suffix_orders = []

    if not isinstance(prefix_orders, list):
        raise ValueError('prefix_orders should be a list if provided')

    if not isinstance(suffix_orders, list):
        raise ValueError('suffix_orders should be a list if provided')

    first = lambda seq: next(iter(seq), UNSIGNED)
    identity = lambda x: x
    default_hash_fun = json.dumps

    if key is None:
        key = identity

    value_types = set(map(type, values))
    value_types.update(set(map(type, prefix_orders)))
    value_types.update(set(map(type, suffix_orders)))
    if len(value_types) > 1:
        raise ValueError('multi types provided in values, prefix_orders, suffix_orders')

    samples = filter(lambda x: x is not UNSIGNED, [first(values)] + [first(prefix_orders)] + [first(suffix_orders)])
    sample = first(samples)

    if sample is UNSIGNED:
        # nothing provided in values, prefix_orders, suffix_orders
        return []

    is_value_type_hashable = True
    try:
        __ = {sample: sample}
    except TypeError:
        is_value_type_hashable = False

    if is_value_type_hashable:
        _hash_fun = hash_fun or identity
    else:
        if hash_fun is not None:
            _hash_fun = hash_fun
        else:
            _hash_fun = default_hash_fun
    hash_fun = _hash_fun

    prefix_orders = list(prefix_orders)
    prefix_orders_set = set(map(hash_fun, prefix_orders))

    if len(prefix_orders) != len(prefix_orders_set):
        raise ValueError('prefix_orders contains duplicated values')

    suffix_orders = list(suffix_orders)
    suffix_orders_set = set(map(hash_fun, suffix_orders))

    if len(suffix_orders) != len(suffix_orders_set):
        raise ValueError('suffix_orders contains duplicated values')

    if prefix_orders_set.intersection(suffix_orders_set):
        raise ValueError('prefix and suffix contains same value')

    order_map = defaultdict(lambda: 1)
    for idx, item in enumerate(prefix_orders):
        order_map[_hash_fun(item)] = idx - len(prefix_orders)

    for idx, item in enumerate(suffix_orders, start=2):
        order_map[_hash_fun(item)] = idx

    sorted_values = sorted(values, key=lambda x: (order_map[hash_fun(x)], key(x)), reverse=reverse)

    return sorted_values


def until(values: Optional[Union[List[Any], Iterable]],
          terminate: Optional[Callable[[Any], bool]] = None,
          default=None) -> Any:
    class Empty:
        pass

    UNSIGNED = Empty()

    if values is None:
        return default

    if terminate is None:
        terminate = lambda v: v is not UNSIGNED

    if isinstance(values, (list, Iterable)):
        for i in values:
            if terminate(i):
                return i
        else:
            pass
        return default
    else:
        raise ValueError('values type should be list, Iterable')



def test_dict_until():
    from pythonic_toolbox.utils.dict_utils import dict_until
    data = {'full_name': 'Albert Lee', 'pen_name': None}
    assert dict_until(data, keys=['name', 'full_name']) == 'Albert Lee'
    assert dict_until(data, keys=['full_name', 'name']) == 'Albert Lee'
    assert dict_until(data, keys=['name', 'english_name']) is None
    assert dict_until(data, keys=['name', 'english_name'], default='anonymous') == 'anonymous'
    # test when pen_name is set None on purpose
    assert dict_until(data, keys=['pen_name'], default='anonymous') is None
    # test when value with None value is not acceptable
    assert dict_until(data, keys=['pen_name'], terminate=lambda x: x is not None, default='anonymous') == 'anonymous'


def test_collect_leaves():
    from pythonic_toolbox.utils.dict_utils import collect_leaves

    assert collect_leaves([]) == []
    assert collect_leaves({}) == []
    assert collect_leaves(None) == []

    # a nested dict-like struct
    my_dict = {
        'node_1': {
            'node_1_1': {
                'node_1_1_1': 'A',
            },
            'node_1_2': {
                'node_1_2_1': 'B',
                'node_1_2_2': 'C',
                'node_1_2_3': None,
            },
            'node_1_3': [  # dict list
                {
                    'node_1_3_1_1': 'D',
                    'node_1_3_1_2': 'E',
                },
                {
                    'node_1_3_2_1': 'FF',
                    'node_1_3_2_2': 'GG',
                }
            ]
        }}

    expected = ['A', 'B', 'C', None, 'D', 'E', 'FF', 'GG']
    assert collect_leaves(my_dict) == expected

    expected = ['A', 'B', 'C', 'D', 'E', 'FF', 'GG']
    assert collect_leaves(my_dict, leaf_pred=lambda lf: lf) == expected

    assert collect_leaves(my_dict, keypath_pred=lambda kp: len(kp) == 1) == []

    expected = ['B', 'C']
    assert collect_leaves(my_dict, keypath_pred=lambda kp: kp[-1] in {'node_1_2_1', 'node_1_2_2'}) == expected

    expected = ['C']
    assert collect_leaves(my_dict, leaf_pred=lambda lf: lf == 'C') == expected
    assert collect_leaves(my_dict,
                          keypath_pred=lambda kp: kp[-1] == 'node_1_2_2',
                          leaf_pred=lambda lf: lf == 'C') == expected

    assert collect_leaves(my_dict,
                          keypath_pred=lambda kp: kp[-1] == 'node_1_1_1',
                          leaf_pred=lambda lf: lf == 'C') == []

    expected = ['D', 'E', 'FF', 'GG']
    assert collect_leaves(my_dict,
                          keypath_pred=lambda kp: len(kp) >= 2 and kp[-2] == 'node_1_3') == expected

    expected = ['FF', 'GG']
    assert collect_leaves(my_dict,
                          keypath_pred=lambda kp: len(kp) >= 2 and kp[-2] == 'node_1_3',
                          leaf_pred=lambda lf: isinstance(lf, str) and len(lf) == 2) == expected


def test_walk_leaves():
    from pythonic_toolbox.utils.dict_utils import walk_leaves

    assert walk_leaves(None) is None
    assert walk_leaves([]) == []
    assert walk_leaves({}) == {}

    assert walk_leaves(None, inplace=True) is None
    assert walk_leaves([], inplace=True) is None
    assert walk_leaves({}, inplace=True) is None

    data = {
        'k1': {
            'k1_1': 1,
            'k1_2': 2,
        },
        'k2': 'N/A',  # stands for not available
    }

    expected = {
        'k1': {
            'k1_1': 2,
            'k1_2': 4,
        },
        'k2': 'N/A',  # stands for not available
    }
    assert walk_leaves(data) == data
    assert walk_leaves(data, trans_fun=lambda x: x * 2 if isinstance(x, int) else x) == expected
    assert walk_leaves(data, trans_fun=lambda x: x * 2 if isinstance(x, int) else x, inplace=True) is None
    assert data == expected

    data = [{'name': 'lml', 'age': 33}, {'name': 'albert', 'age': 18}]
    expected = [{'name': 'lml', 'age': 66}, {'name': 'albert', 'age': 36}]
    assert walk_leaves(data, trans_fun=lambda x: x * 2 if isinstance(x, int) else x) == expected
    assert walk_leaves(data, trans_fun=lambda x: x * 2 if isinstance(x, int) else x, inplace=True) is None
    assert data == expected


def test_dict_obj():
    import pytest
    from pythonic_toolbox.utils.dict_utils import DictObj

    naive_dct = {
        'key1': 'val1',
        'key2': 'val2',
    }

    obj = DictObj(naive_dct)
    # same behavior like ordinary dict according to the python version (FILO for popitem for 3.6+)
    assert obj.popitem() == ('key2', 'val2')
    assert obj.popitem() == ('key1', 'val1')
    with pytest.raises(KeyError) as __:
        obj.popitem()
    obj.key3 = 'val3'
    assert obj.pop('key3', None) == 'val3'
    assert obj.pop('key4', None) is None

    person_dct = {'name': 'Albert', 'age': '34', 'sex': 'Male', 'languages': ['Chinese', 'English']}

    person = DictObj(person_dct)
    assert person.to_dict() == person_dct
    assert set(person.keys()) == {'name', 'age', 'sex', 'languages'}
    assert hasattr(person, 'name') is True
    assert person.name == 'Albert'
    assert person['name'] == 'Albert'
    person.languages.append('Japanese')
    assert person.languages == ['Chinese', 'English', 'Japanese']

    person.height = '170'
    assert person['height'] == '170'
    assert 'height' in person
    assert 'height' in person.keys()
    assert hasattr(person, 'height') is True
    del person['height']
    assert 'height' not in person
    assert 'height' not in person.keys()
    person['height'] = '170cm'

    person.update({'weight': '50'})
    weight_val = person.pop('weight')
    assert weight_val == '50'
    person.update(DictObj({'weight': '50kg'}))
    assert person.weight == '50kg'

    expected = {
        'name': 'Albert', 'age': '34', 'sex': 'Male',
        'languages': ['Chinese', 'English', 'Japanese'],  # appended new language
        'height': '170cm',  # new added attribute
        'weight': '50kg',  # new added attribute
    }
    assert person.to_dict() == expected

    repr_expected: str = ("{'name': 'Albert', 'age': '34', 'sex': 'Male', "
                          "'languages': ['Chinese', 'English', 'Japanese'],"
                          " 'height': '170cm', 'weight': '50kg'}")
    assert repr(person) == repr_expected


def test_final_dict_obj():
    import pytest
    from pythonic_toolbox.utils.dict_utils import FinalDictObj

    person_dct = {'name': 'Albert', 'age': '34', 'sex': 'Male', 'languages': ['Chinese', 'English']}

    fixed_person = FinalDictObj(person_dct)
    assert fixed_person.name == 'Albert'

    with pytest.raises(RuntimeError) as exec_info:
        fixed_person.name = 'Steve'
    expected_error_str = 'Not allowed to assign attribute name with value Steve for an initialized FinalDictObj'
    assert exec_info.value.args[0] == expected_error_str

    with pytest.raises(RuntimeError) as __:
        fixed_person.popitem()

    with pytest.raises(RuntimeError) as __:
        fixed_person.pop('name')

    assert type(fixed_person.languages) == tuple
    with pytest.raises(AttributeError) as exec_info:
        # list values are changed into tuple to avoid being modified
        fixed_person.languages.append('Japanese')
    expected_error_str = "'tuple' object has no attribute 'append'"
    assert exec_info.value.args[0] == expected_error_str
    assert fixed_person.to_dict() == person_dct



from collections import UserDict
from typing import Any, Callable, Optional, List, Hashable, Union
from copy import deepcopy


def dict_until(obj, keys: list, terminate: Optional[Callable[[Any], bool]] = None, default=None):
    class Empty:
        pass

    UNSIGNED = Empty()

    if terminate is None:
        terminate = lambda v: v is not UNSIGNED

    from pythonic_toolbox.utils.list_utils import until

    # default is for value
    val = default
    key = until(keys, lambda k: terminate(obj.get(k, UNSIGNED)), default=UNSIGNED)
    if key is not UNSIGNED:
        val = obj[key]
    return val


def collect_leaves(data: Optional[Union[dict, List]] = None,
                   keypath_pred: Optional[Callable[[List[Hashable]], bool]] = None,
                   leaf_pred: Optional[Callable[[Any], bool]] = None) -> List[Any]:
    leaves = list()
    if not data:
        return leaves

    keypath_pred_comb = lambda x: keypath_pred is None or keypath_pred(x)
    leaf_pred_comb = lambda x: leaf_pred is None or leaf_pred(x)

    def _traverse(_data, keypath=None):
        keypath = keypath or []
        if isinstance(_data, dict):
            return {k: _traverse(v, keypath + [k]) for k, v in _data.items()}
        elif isinstance(_data, list):
            return [_traverse(elem, keypath) for elem in _data]
        else:
            # no container, just values (str, int, float, None, obj etc.)
            if keypath_pred_comb(keypath) and leaf_pred_comb(_data):
                leaves.append(_data)

            return _data

    _traverse(data)
    return leaves


def walk_leaves(data: Optional[Union[dict, List]] = None,
                trans_fun: Optional[Callable[[Any], Any]] = None,
                inplace=False) -> Optional[Union[dict, List]]:
    """
    :param data: data can be nested dict, list
    :param trans_fun: leaf transform function
    :param inplace: change values in place or not
    :return: replace data with transformed leaves, will return None in transform inplace
    """
    if data is None:
        return data
    if not isinstance(data, (dict, list)):
        raise ValueError('data must be dict or list')

    if inplace is True:
        obj = data
    else:
        # won't touch the original data
        obj = deepcopy(data)

    if trans_fun is None:
        return obj if inplace is False else None

    def _traverse(_obj, parent: Optional[Union[dict, list]] = None,
                  idx: Optional[Union[int, Hashable]] = None) -> None:
        """
        This inner function transform leaves value inplace
        """
        if isinstance(_obj, dict):
            __ = {k: _traverse(v, _obj, k) for k, v in _obj.items()}
        elif isinstance(_obj, list):
            __ = [_traverse(elem, _obj, idx) for idx, elem in enumerate(_obj)]
        else:
            # no container, just values (str, int, float, None,  obj etc.)
            parent[idx] = trans_fun(_obj)

    _traverse(obj)
    return obj if inplace is False else None


class DictObj(UserDict):
    __lst_factory = list

    def __init__(self, in_dict: dict):

        in_dict = deepcopy(in_dict)

        if any(map(lambda key: not isinstance(key, str), in_dict.keys())):
            raise ValueError('Only string key dict are allowed for DictObj/FinalDictObj input dict')

        for key, val in in_dict.items():
            if isinstance(val, list):
                lst_factory = object.__getattribute__(self, f'_{self.__class__.__name__}__lst_factory')
                in_dict[key] = lst_factory(self.__class__(x) if isinstance(x, dict) else x for x in val)
            elif isinstance(val, dict):
                in_dict[key] = self.__class__(val)
            else:
                in_dict[key] = val

        super(DictObj, self).__init__(**in_dict)

    def __setitem__(self, key, item):
        if isinstance(item, list):
            lst_factory = object.__getattribute__(self, f'_{self.__class__.__name__}__lst_factory')
            self.data[key] = lst_factory(self.__class__(x) if isinstance(x, dict) else x for x in item)
        elif isinstance(item, dict):
            self.data[key] = self.__class__(item)
        else:
            self.data[key] = item

    def popitem(self):
        """
        Override popitem from MutableMapping, make behavior popitem FILO like ordinary dict since 3.6
        """
        return self.data.popitem()

    def __getattribute__(self, item):
        if item == 'data':
            return self.__dict__['data']
        else:
            return super(DictObj, self).__getattribute__(item)

    def __delitem__(self, key):
        del self.data[key]

    def __setattr__(self, key, value):
        """DictObj that cannot change attribute"""
        if key == 'data':
            object.__setattr__(self, 'data', value)
        else:
            data = object.__getattribute__(self, 'data')
            if isinstance(value, list):
                lst_factory = object.__getattribute__(self, f'_{self.__class__.__name__}__lst_factory')
                if lst_factory == tuple:
                    print('Gotcha')
                data[key] = lst_factory(self.__class__(x) if isinstance(x, dict) else x for x in value)
            elif isinstance(value, dict):
                data[key] = self.__class__(value)
            else:
                data[key] = value

            object.__setattr__(self, 'data', data)

    def __getattr__(self, item):
        try:
            data = self.__dict__['data']
            res = data[item]
        except KeyError:
            raise AttributeError(f'AttributeError {item}')
        else:
            return res

    def __delattr__(self, item):
        del self[item]

    def to_dict(self):
        result = {}
        for key, item in self.data.items():
            if isinstance(item, (list, tuple)):
                result[key] = [x.to_dict() if isinstance(x, DictObj) else x for x in item]
            elif isinstance(item, DictObj):
                result[key] = item.to_dict()
            else:
                result[key] = item
        return result


class FinalDictObj(DictObj):
    __lst_factory = tuple
    __is_frozen = False

    def __init__(self, in_dict: dict):

        in_dict = deepcopy(in_dict)

        super(FinalDictObj, self).__init__(in_dict)
        self._freeze()

    def _freeze(self):
        self.__is_frozen = True

    def __setitem__(self, key, value):
        """DictObj that cannot change attribute"""
        if self.__is_frozen is True:
            raise RuntimeError(f'Cannot set attribute/item {key} in FinalDictObj')
        super(FinalDictObj, self).__setitem__(key, value)

    def __delitem__(self, key):
        if self.__is_frozen is True:
            raise RuntimeError(f'Cannot del attribute or item {key} in an initialized FinalDictObj')
        super(FinalDictObj, self).__delitem__(key)

    def popitem(self):
        """
        Override popitem from MutableMapping, make behavior popitem FILO like ordinary dict since 3.6
        """
        if self.__is_frozen is True:
            raise RuntimeError(f'Cannot popitem attribute/item in initialed FinalDictObj')
        return super(FinalDictObj, self).popitem()

    def __setattr__(self, key, value):
        """DictObj that cannot change attribute"""
        if key == '_FinalDictObj__is_frozen' and value is True:
            # __is_frozen can only be assigned once
            object.__setattr__(self, key, value)
        else:
            if self.__is_frozen:
                raise RuntimeError(
                    f"Not allowed to assign attribute {key} with value {value} for an initialized FinalDictObj")

            super(FinalDictObj, self).__setattr__(key, value)

    def __delattr__(self, item):
        if self.__is_frozen is True:
            raise RuntimeError(f'Cannot del attribute or item {item} in an initialized FinalDictObj')
        super(FinalDictObj, self).__delattr__(item)

    def update(self, *args, **kwargs):
        if self.__is_frozen is True:
            raise RuntimeError(f'Update not allowed for an initialized FinalDictObj')
        super(FinalDictObj, self).update(*args, **kwargs)



from itertools import count
import pytest

from pythonic_toolbox.utils.list_utils import until, sort_with_custom_orders


def test_until():
    counter = count(1, 2)  # generator of odd numbers: 1, 3, 5, 7 ...
    assert until([], default=3) == 3  # nothing provided, return default
    assert until(counter, lambda x: x > 10) == 11
    assert until([1, 2, 3], lambda x: x > 10, default=11) == 11
    assert until(None, lambda x: x > 10, default=11) == 11


def test_sort_with_custom_orders():
    assert sort_with_custom_orders([]) == []
    assert sort_with_custom_orders([], prefix_orders=[], suffix_orders=[]) == []
    assert sort_with_custom_orders([], prefix_orders=['master']) == []

    values = ['branch2', 'branch1', 'branch3', 'master', 'release']
    expected = ['master', 'release', 'branch1', 'branch2', 'branch3']
    assert sort_with_custom_orders(values, prefix_orders=['master', 'release']) == expected
    assert sort_with_custom_orders(values, prefix_orders=['master', 'release'], reverse=True) == expected[::-1]

    values = [1, 2, 3, 9, 9]
    expected = [9, 9, 1, 2, 3]
    assert sort_with_custom_orders(values, prefix_orders=[9, 8, 7]) == expected

    values = [1, 2, 3, 9]
    expected = [9, 2, 3, 1]
    assert sort_with_custom_orders(values, prefix_orders=[9], suffix_orders=[1]) == expected

    # testing for unshashable values
    values = [[2, 2], [1, 1], [3, 3], [6, 0]]
    expected = [[3, 3], [1, 1], [2, 2], [6, 0]]
    assert sort_with_custom_orders(values, prefix_orders=[[3, 3]], key=lambda x: sum(x)) == expected

    values = [{2: 2}, {1: 1}, {1: 2}]
    expected = [{2: 2}, {1: 1}, {1: 2}]
    assert sort_with_custom_orders(values, prefix_orders=[{2: 2}],
                                   key=lambda data: sum(data.values())) == expected
    assert sort_with_custom_orders(values, prefix_orders=[{2: 2}],
                                   key=lambda data: sum(data.values()), hash_fun=tuple) == expected

    with pytest.raises(ValueError) as exec_info:
        sort_with_custom_orders([1, 2, 3], prefix_orders=[3], suffix_orders=[3])
    assert exec_info.value.args[0] == 'prefix and suffix contains same value'

    with pytest.raises(ValueError) as exec_info:
        sort_with_custom_orders([1, 2, 3], prefix_orders=[1, 1])
    assert exec_info.value.args[0] == 'prefix_orders contains duplicated values'

    class Person:
        def __init__(self, id, name, age):
            self.id = id
            self.name = name
            self.age = age

        def __lt__(self, other: 'Person'):
            return self.age < other.age

        def __eq__(self, other: 'Person'):
            return self.age == other.age

        def __hash__(self):
            return self.id

        def __str__(self):
            return f'Person({self.id}, {self.name}, {self.age})'

        def __repr__(self):
            return str(self)

    Albert = Person(1, 'Albert', 28)
    Alice = Person(2, 'Alice', 26)
    Menglong = Person(3, 'Menglong', 33)

    persons = [Albert, Alice, Menglong]
    expected = [Alice, Albert, Menglong]
    assert sort_with_custom_orders(persons) == expected

    expected = [Menglong, Alice, Albert]
    assert sort_with_custom_orders(persons, prefix_orders=[Menglong, Person(4, 'Anyone', 40)]) == expected



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

long_description = 'A toolbox' \
                   ' containing multi useful pythonic utils, functions, decorators etc.'

url = 'https://github.com/albertmenglongli/pythonic-toolbox'

setup(
    name='pythonic-toolbox',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.1.9',

    description='a toolbox with pythonic utils, tools',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/albertmenglongli/pythonic-toolbox',

    # Author details
    author='menglong.li',
    author_email='albert.menglongli@gmail.com',

    # Choose your license
    license='Apache2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    # What does your project relate to?
    keywords='toolbox',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    py_modules=["pythonic_toolbox"],

    install_requires=[
        'pytest'
    ]

)
